unit XAsync;

(**************************************************************************
 *     This unit is part of 2labs Communication components.               *
 *     Copyright (C) 2001 Ondrej Urik. All rights reserved.               *
 *                                                                        *
 *     NOTE                                                               *
 *     This unit may be used with other communication units/packages.     *
 *     Modifying allowed if the above copyright is included.              *
 **************************************************************************)

(***   Version 1.0                                                      ***)

{$H+,R-,B-}

interface

uses
  Windows, SysUtils;

resourcestring
  SHInvalidHandle = 'Invalid HASYNC handle value.';

type
  HASYNC = type THandle;
  TAsyncProc = procedure (Success: Boolean; Data: Pointer; Count: Longint) of object;
  TWaitProc = function: Boolean of object;

  EAsyncError = class(Exception);

function IsHandleValid(Handle: HASYNC): Boolean;
function IsAsyncDone(Handle: HASYNC): Boolean;

function InternalInitAsync(Handle: THandle; AsyncProc: TAsyncProc;
  AutoClose: Boolean): HASYNC;
function InternalCloseAsync(Handle: HASYNC): Boolean;

function InternalReadAsync(Handle: HASYNC; var Data; Count: DWORD): DWORD;
function InternalReadStringAsync(Handle: HASYNC; var Data: string): DWORD;
function InternalWriteAsync(Handle: HASYNC; const Data; Count: DWORD): DWORD;
function InternalWaitAsync(Handle: HASYNC; Process: TWaitProc): Boolean;


implementation

uses
  Forms, Classes;

type
  AOptions = set of (asDone, asSuccess, asFree, asString);
  PAsync = ^TAsync;
  TAsync = record
    Handle: THandle;  //Handle to device
    HHandle: THandle; //Internal handle
    Buffer: Pointer;
    Count: DWORD;
    Overlapped: TOverlapped;
    AsyncProc: TAsyncProc;
    Options: AOptions;
  end;

  TAsyncThread = class(TThread)
  private
    FAsync: PAsync;
  protected
    procedure DoAsyncProc;
    procedure Execute; override;
  public
    constructor Create(Handle: HASYNC); virtual;
  end;

{TAsyncThread}
constructor TAsyncThread.Create(Handle: HASYNC);
begin
  inherited Create(True);
  FreeOnTerminate := True;
  FAsync := PAsync(Handle);
  Exclude(FAsync^.Options, asDone);
end;

procedure TAsyncThread.DoAsyncProc;
begin
  with FAsync^ do
    AsyncProc(asSuccess in Options, Buffer, Count);
end;

procedure TAsyncThread.Execute;
var
  nCount: DWORD;
  PStr: PString;
begin
  Application.Title := 'Started';
  with FAsync^ do
  begin
    WaitForSingleObject(Overlapped.hEvent, INFINITE);
    if GetOverlappedResult(Handle, Overlapped, nCount, False) then
      Include(Options, asSuccess);
    Count := nCount;
    if asString in Options then
    begin
      PStr := Buffer;
      SetLength(PStr^, Count);
    end;
    CloseHandle(Overlapped.hEvent);
    if Assigned(AsyncProc) then
      Synchronize(DoAsyncProc);
    Include(Options, asDone);
    if asFree in Options then InternalCloseAsync(THandle(FAsync));
  end;
end;

{Internal}
procedure InternalStartThread(Handle: HASYNC);
var AT: TAsyncThread;
begin
  AT := TAsyncThread.Create(Handle);
  AT.Resume;
end;

function IsHandleValid(Handle: HASYNC): Boolean;
begin
  Result := (Handle<>0) and (PAsync(Handle)^.HHandle=$FFFFFFFF-Handle);
end;

function IsAsyncDone(Handle: HASYNC): Boolean;
begin
  Result := not IsHandleValid(Handle) or (asDone in PAsync(Handle)^.Options);
end;

procedure InternalCheckHandle(Handle: HASYNC);
begin
  if not IsHandleValid(Handle) then raise EAsyncError.Create(SHInvalidHandle);
end;

function InternalInitAsync(Handle: THandle; AsyncProc: TAsyncProc;
  AutoClose: Boolean): HASYNC;
var P: ^TAsync;
begin
  GetMem(P, SizeOf(TAsync));
  try
    Result := THandle(P);
    P^.Handle := Handle;
    P^.AsyncProc := AsyncProc;
    with P^ do
    begin
      Buffer := nil;
      Count := 0;
      HHandle := $FFFFFFFF-Result;
      Options := [asDone];
      if AutoClose then Include(Options, asFree);
      FillChar(Overlapped, SizeOf(Overlapped), 0);
      Overlapped.hEvent := CreateEvent(nil, True, True, nil);
    end;
  except
    FreeMem(P, SizeOf(TAsync));
    raise;
  end;
end;

function InternalCloseAsync(Handle: HASYNC): Boolean;
var P: PAsync;
begin
  InternalCheckHandle(Handle);
  P := PAsync(Handle);
  Result := (asDone in P^.Options);
  if Result then
  begin
    P^.HHandle := 0;
    FreeMem(P, SizeOf(TAsync));
  end;
end;

function InternalReadStringAsync(Handle: HASYNC; var Data: string): DWORD;
var
  ComStat: TComStat;
  P: PAsync;
begin
  InternalCheckHandle(Handle);
  P := PAsync(Handle);
  if ClearCommError(P^.Handle, Result, @ComStat) then
    SetLength(Data, ComStat.cbInQue);
  P^.Buffer := @Data;
  ReadFile(P^.Handle, Data[1], ComStat.cbInQue, Result, @P^.Overlapped);
  Include(P^.Options, asString);
  InternalStartThread(Handle);
end;

function InternalReadAsync(Handle: HASYNC; var Data; Count: DWORD): DWORD;
var P: PAsync;
begin
  InternalCheckHandle(Handle);
  P := PAsync(Handle);
  P^.Buffer := @Data;
  ReadFile(P^.Handle, Data, Count, Result, @P^.Overlapped);
  InternalStartThread(Handle);
end;

function InternalWriteAsync(Handle: HASYNC; const Data; Count: DWORD): DWORD;
var P: PAsync;
begin
  InternalCheckHandle(Handle);
  P := PAsync(Handle);
  WriteFile(P^.Handle, Data, Count, Result, @P^.Overlapped);
  InternalStartThread(Handle);
end;

function InternalWaitAsync(Handle: HASYNC; Process: TWaitProc): Boolean;
begin
  InternalCheckHandle(Handle);
  with PAsync(Handle)^ do
  begin
    Result := asFree in Options;
    Exclude(Options, asFree);
    while not (asDone in Options) do
    begin
      Application.ProcessMessages;
      if Assigned(Process) and Process then
      begin
        if Result then
          Include(Options, asFree);
        Result := False;
        Break;
      end;
    end;
  end;
  if Result then
    InternalCloseAsync(Handle)
  else
    Result := asDone in PAsync(Handle)^.Options;
end;

end.
